/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                INTEL CORPORATION PROPRIETARY INFORMATION                //
//   This software is supplied under the terms of a license agreement or   //
//  nondisclosure agreement with Intel Corporation and may not be copied   //
//   or disclosed except in accordance with the terms of that agreement.   //
//       Copyright (c) 2003 Intel Corporation. All Rights Reserved.        //
//                                                                         //
//  File:      pnlContextPersistence.hpp                                   //
//                                                                         //
//  Purpose:   Persistence contexts                                        //
//                                                                         //
//  Author(s):                                                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////


#ifndef __PNLCONTEXTPERSISTENCE_HPP__
#define __PNLCONTEXTPERSISTENCE_HPP__

#pragma warning (push, 3 )
#include <fstream>
#include <sstream>
#pragma warning(pop)

#include "pnlContext.hpp"

PNL_BEGIN

class PNL_API CContextPersistence: public CContext
{
public:
    CContextPersistence() {}
    bool SaveAsXML(const std::string &filename) const;
    bool LoadXML(const std::string &filename);
};


class CContextSave: public CContext
{
public:
    CContextSave(const std::string &filename): m_File(filename.c_str()) {}

    virtual void BeginTraverseObject(const std::string &typeName, TreeEntry& rEntry);
    virtual void   EndTraverseObject(const std::string &typeName, TreeEntry& rEntry);
    virtual void DoEndTraverseObject() = 0;
    void AddAttribute(const char *attr, const char *value)
    {
        m_aAttrName.push_back(std::string(attr));
        m_aValue.push_back(std::string(value));
    }
    void AddAttribute(const char *attr, const int value)
    {
	char buf[20];

	sprintf(buf, "%i", value);
	AddAttribute(attr, buf);
    }
    void AddAttribute(const char *attr, bool value)
    {
        AddAttribute(attr, value ? "1":"0");
    }
    void AddAttribute(const char *attr, const float value)
    {
	char buf[40];

	sprintf(buf, "%lg", (double)value);
	AddAttribute(attr, buf);
    }
    void AddText(const char *text)
    {
        m_Text.append(text);
    }

    void PlanForWriting();
    virtual void BeforeInterior() {}
    virtual ~CContextSave();

protected:
    bool IsPlanned() const { return m_bPlanned; }

protected:
    pnlVector<std::string> m_aTag;
    std::string m_Text;
    pnlVector<std::string> m_aAttrName;
    pnlVector<std::string> m_aValue;
    std::ofstream m_File;
    bool m_bPlanned;
};

class CContextSaveXML: public CContextSave
{
public:
    CContextSaveXML(const std::string &filename);
    ~CContextSaveXML();

    virtual void BeginTraverseObject(const std::string &typeName, TreeEntry& rEntry);
    virtual void DoEndTraverseObject();
    virtual void BeforeInterior();
};

class CContextLoad: public CContext
{
public:
    CContextLoad(const std::string &filename): m_File(filename.c_str()) {}
    virtual void BeginTraverseObject(const std::string &typeName, TreeEntry& rEntry);
    virtual void   EndTraverseObject(const std::string &typeName, TreeEntry& rEntry);
    struct TreeTextEntry
    {
        TreeTextEntry(const std::string &name): m_Name(name), m_iChildren(-1) {}
        TreeTextEntry(): m_iChildren(-1) {}
        const TreeTextEntry &operator=(const TreeTextEntry &a)
        {
            if(this != &a)
            {
                m_Text = a.m_Text;
                m_Name = a.m_Name;
                m_aAttrName = a.m_aAttrName;
                m_aAttrValue = a.m_aAttrValue;
                m_iChildren = a.m_iChildren;
            }
            return *this;
        }
        std::string m_Text;
        std::string m_Name;
        pnlVector<std::string> m_aAttrName;
        pnlVector<std::string> m_aAttrValue;
        int m_iChildren;
    };

    ~CContextLoad()
    {
        m_File.close();
    }

    virtual bool HandleObjectBeforeInterior() const { return false; }

    void RecursiveCopying(int iaaEntryTxt);

    void GetText(std::string &text)
    {
        text = ((TreeTextEntry*)Current().m_pUser)->m_Text;
    }

    void GetAttribute(std::string &attrValue, const std::string &attrName)
    {
        TreeTextEntry *pEnt = (TreeTextEntry*)Current().m_pUser;

        FindAttribute(*pEnt, attrValue, attrName);
    }
    void GetAttribute(bool *attrValue, const std::string &attrName)
    {
	std::string attrValueStr;

	GetAttribute(attrValueStr, attrName);
	attrValue[0] = (attrValueStr[0] == '1');
    }
    void GetAttribute(int *attrValue, const std::string &attrName)
    {
	std::string attrValueStr;

	GetAttribute(attrValueStr, attrName);
	attrValue[0] = atoi(attrValueStr.c_str());
    }
    void GetAttribute(float *attrValue, const std::string &attrName)
    {
	std::string attrValueStr;

	GetAttribute(attrValueStr, attrName);
	attrValue[0] = (float)atof(attrValueStr.c_str());
    }

protected:
    void FindAttribute(TreeTextEntry &rEnt, std::string &attrValue, const std::string &attrName)
    {
        for(int i = 0; i < rEnt.m_aAttrName.size(); ++i)
        {
            if(rEnt.m_aAttrName[i] == attrName)
            {
                attrValue = rEnt.m_aAttrValue[i];
                return;
            }
        }

        attrValue.resize(0);
    }

    virtual void GetObjWithObjTypeName(CPNLBase **ppObj,
        std::string *pObjTypeName, const TreeEntry &rEntry)
    {
        *ppObj = 0;
        FindAttribute(*((TreeTextEntry*)rEntry.m_pUser), *pObjTypeName,
            std::string("TypeName"));
    }


protected:
    std::ifstream m_File;
    pnlVector<pnlVector<TreeTextEntry> > m_aaEntryTxt;
};


class CContextLoadXML: public CContextLoad
{
public:
    CContextLoadXML(const std::string &filename): CContextLoad(filename) {}

    class XMLLeaf
    {
    public:
        XMLLeaf(pnlVector<pnlVector<TreeTextEntry> >& aaEntry, int major, int minor)
            : m_paaEntry(&aaEntry), m_Major(major), m_Minor(minor)
        {}
        
        TreeTextEntry& operator()()
        {
            return (*m_paaEntry)[m_Major][m_Minor];
        }

    private:
        pnlVector<pnlVector<TreeTextEntry> > *m_paaEntry;
        int m_Major;
        int m_Minor;
    };

    virtual void BeginTraverseObject(const std::string &typeName, TreeEntry& rEntry);
    bool SwallowXML();

};

PNL_END

#endif // include guard
