/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                INTEL CORPORATION PROPRIETARY INFORMATION                //
//   This software is supplied under the terms of a license agreement or   //
//  nondisclosure agreement with Intel Corporation and may not be copied   //
//   or disclosed except in accordance with the terms of that agreement.   //
//       Copyright (c) 2003 Intel Corporation. All Rights Reserved.        //
//                                                                         //
//  File:      pnlMlStaticStructLearnHC.cpp                                //
//                                                                         //
//  Purpose:   Implementation of the CMlStaticStructLearnHC class          //
//                                                                         //
//  Author(s):                                                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////
#include "pnlConfig.hpp"
#include "pnlMlStaticStructLearnHC.hpp"

#include "mpiWrap.h" 
#include "wtime.h"

#ifdef _OPENMP
#include <omp.h>
#endif

PNL_USING

CMlStaticStructLearnHC::CMlStaticStructLearnHC( CStaticGraphicalModel* pGrModel, 
						  ELearningTypes LearnType, 
						  EOptimizeTypes AlgorithmType, 
						  EScoreFunTypes ScoreType, 
						  int nMaxFanIn,
						  intVector& vAncestor,
						  intVector& vDescent,
						  int nRestarts ):
							CMlStaticStructLearn( pGrModel, LearnType, 
								                   AlgorithmType,ScoreType,
												   nMaxFanIn,vAncestor,vDescent )
{
	m_nRestarts = nRestarts;
	m_minProgress = (float)1e-4;
	m_nSteps = 100;
}

CMlStaticStructLearnHC* CMlStaticStructLearnHC::Create( CStaticGraphicalModel* pGrModel, 
						  ELearningTypes LearnType, 
						  EOptimizeTypes AlgorithmType, 
						  EScoreFunTypes ScoreType, 
						  int nMaxFanIn,
						  intVector& vAncestor,
						  intVector& vDescent,
						  int nRestarts )
{
	CMlStaticStructLearnHC* ret = new CMlStaticStructLearnHC( pGrModel, LearnType, 
								                   AlgorithmType,ScoreType,
												   nMaxFanIn, vAncestor,
												   vDescent, nRestarts );
	return ret;
}

void CMlStaticStructLearnHC::SetMinProgress(float minProgress)
{
	m_minProgress = minProgress;
}

void CMlStaticStructLearnHC::SetMaxSteps(int steps)
{
	m_nSteps = steps;
}

void CMlStaticStructLearnHC::Learn()
{
	if(m_Vector_pEvidences.size() == 0)
		PNL_THROW(CInconsistentState, "should set the data first");
	
	CGraph* iGraph = m_pGrModel->GetGraph();
	CDAG* iDAG = CDAG::Create(*iGraph);
	CDAG* pDAG;
	CDAG* pBestDAG = NULL;
	float BestScore = (float)-1e37;
	int irestarts = m_nRestarts;
	int i, istart;
	float score;

	for(istart=0; istart<irestarts; istart++)
	{
		if(istart>0)
		{
			delete iDAG;
			intVector	vDiscrete, vContinuous;
			const CNodeType* nt;
			for(i=0; i<m_nNodes; i++)
			{
				nt = m_pGrModel->GetNodeType(i);
				if( nt->IsDiscrete() )
					vDiscrete.push_back(i);
				else
					vContinuous.push_back(i);
			}
			iDAG = CDAG::RandomCreateADAG(m_nNodes, m_nMaxFanIn, vDiscrete, vContinuous);
		}
		LearnInOneStart(iDAG, &pDAG, &score);
		if(score > BestScore)
		{
			delete pBestDAG;
			pBestDAG = pDAG->Clone();
			BestScore = score;
		}
		//delete pDAG;  //lai, 05-17-2004
	}
	//delete iDAG;  //lai, 05-17-2004
	m_pResultDAG = pBestDAG->Clone();
	m_critValue.push_back(BestScore);
	delete pBestDAG;
}

#ifndef _MPI

#define REVERSE_EDGES_CHUNK  256  //added by lai, 06-09-2004
typedef struct {                  //added by lai, 06-09-2004
  int   parentNode;
  int   childNode;
  float semi_scoreimprove;
} EDGEREVERSE;                    //added by lai, 06-09-2004

EDGEOP      *move_global;                                            //added by lai, 05-17-2004
float       *max_improve_global, *max_start_global, *max_end_global; //added by lai, 05-17-2004
int         *newFamily_global, *indexes_global, *Parents_Buffer_global;  //added by lai, 05-17-2004
#pragma omp threadprivate(move_global)  //added by lai, 05-17-2004
#pragma omp threadprivate(max_improve_global, max_start_global, max_end_global) //added by lai, 05-17-2004
#pragma omp threadprivate(newFamily_global) //added by lai, 05-17-2004
#pragma omp threadprivate(indexes_global, Parents_Buffer_global) //added by lai, 05-17-2004

void CMlStaticStructLearnHC::LearnInOneStart(CDAG* InitDag, 
					CDAG** LearnedDag, float* LearnedScore)
{
	int		j, step=0;
	bool	progress = true;
	CDAG*	iDAG = InitDag->Clone();
	floatVector FamilyScore;
	float init_score = ScoreDAG(iDAG, &FamilyScore);	// Eric

	int			nValidMoves;
	EDGEOPVECTOR vValidMoves;
	float		best_score, score_gate;
	intVector	vAncestor, vDescent;
	intVector	vDiscrete, vContinuous;
	const CNodeType* nt;
	for(j=0; j<m_nNodes; j++)
	{
		nt = m_pGrModel->GetNodeType(j);
		if( nt->IsDiscrete() )
			vDiscrete.push_back(j);
		else
			vContinuous.push_back(j);
	}
	vAncestor.assign(m_vAncestor.begin(), m_vAncestor.end());
	vDescent.assign(m_vDescent.begin(), m_vDescent.end());

#pragma omp parallel
{
	newFamily_global   = new int [m_nMaxFanIn + 1];
	indexes_global     = new int [m_nMaxFanIn];
	Parents_Buffer_global  = new int [m_nMaxFanIn];
}

	int n_reverseslimit = REVERSE_EDGES_CHUNK;    //lai, 06-09-2004
	int n_totalreverses;             //lai, 06-09-2004
    EDGEREVERSE* ReverseEdges = (EDGEREVERSE *)malloc(n_reverseslimit * sizeof(EDGEREVERSE)); //lai, 06-09-2004

	timer_start(2);  //lai, 04-20-2004
	//VTResume();
    while ( step<m_maxIterIPF && progress )
	//while ( step<m_nSteps && progress )
	{
		nValidMoves = 0;
		n_totalreverses = 0;             //lai, 06-09-2004
		float score_improve, score_start, score_end;  //lai, 05-11-2004
		int k;

#ifndef _OPENMP
		EDGEOP  move_content;
		move_content.originalEdge.startNode =       0;
        move_content.originalEdge.endNode   =       0;
		move_content.DAGChangeType          = DAG_ADD;
		float m_start_content, m_end_content;         //lai, 05-11-2004
		float m_improve_content = 0.0f;               //lai, 05-11-2004
		move_global = &move_content;
        max_improve_global = &m_improve_content;
		max_start_global   = &m_start_content;
		max_end_global     = &m_end_content;
#else
		int num_threads = omp_get_max_threads();
		EDGEOP *moves_content = new EDGEOP[num_threads];
		float *m_starts_content = new float[num_threads];
		float *m_ends_content = new float[num_threads];
		float *m_improves_content = new float[num_threads];
        for(k = 0; k < num_threads; k ++) {
			moves_content[k].originalEdge.startNode =       0;
			moves_content[k].originalEdge.endNode   =       0;
			moves_content[k].DAGChangeType          = DAG_ADD;
			m_improves_content[k] = 0.0f;
		}
#pragma omp parallel
		{
			move_global = moves_content + omp_get_thread_num();
			max_improve_global = m_improves_content + omp_get_thread_num();
			max_start_global   = m_starts_content + omp_get_thread_num();
			max_end_global     = m_ends_content + omp_get_thread_num();
		}
#endif

		//printf("\nbegin process moves at %f\n", (float)(clock()-start_GetMoves)/CLOCKS_PER_SEC);

#pragma omp parallel for schedule(static) private(score_improve, k) reduction(+: nValidMoves)
		for(j=0; j<m_nNodes; j++) {
		    for(int i=0; i<m_nNodes; i++)
			{
			  //newFamily->clear();     //lai, 05-17-2004
              int position, nParrents;
		      if(i == j)
			  {
			   continue;
			  }
              int nDirect = iDAG->GetEdgeDirect(i, j);
			  if(nDirect == 1)  // try remove, change
			  {
			    //if((!bCheck || CheckValid(*pvAncesstorVector, *pvDescendantsVector, eOP)) && (!bCheck2 || CheckValid2(*pvNotParents, *pvNotChild, eOP)))
				nParrents = iDAG->GetParents(j, newFamily_global); //lai, 05-17-2004
                position = nParrents;
                for( k = 0; k < nParrents; k ++)
					if (newFamily_global[k] == i) {
						position = k;
						break;
					}
				for(k = position; k < nParrents - 1; k ++)
					newFamily_global[k] = newFamily_global[k + 1];
				newFamily_global[nParrents - 1] = j;
				score_improve = ScoreFamily(nParrents, newFamily_global, indexes_global) - FamilyScore[j]; //lai, 05-17-2004
				if (score_improve > *max_improve_global) { 
					*max_improve_global = score_improve;   
					move_global->originalEdge.startNode =       i; 
					move_global->originalEdge.endNode   =       j; 
					move_global->DAGChangeType          = DAG_DEL; 
				}                                  
				nValidMoves++;

//#ifndef _OPENMP
				if(iDAG->IsValidMove(i, j, DAG_REV, Parents_Buffer_global)) { //lai, 05-17-2004
				    nParrents = iDAG->GetParents(i, newFamily_global); //add an edge
					newFamily_global[nParrents] = j;
					if((nParrents + 1) <= m_nMaxFanIn) { //lai, 05-17-2004
//*****
#ifndef _OPENMP
					  newFamily_global[nParrents + 1] = i;
				      score_end = score_improve;
					  score_start = ScoreFamily(nParrents + 2, newFamily_global, indexes_global) - FamilyScore[i]; //lai, 05-17-2004
					  score_improve += score_start;
					  if (score_improve > *max_improve_global) { 
					      *max_improve_global = score_improve;   
					      *max_start_global   =   score_start;   
					      *max_end_global     =     score_end;   
					      move_global->originalEdge.startNode =       i; 
					      move_global->originalEdge.endNode   =       j; 
					      move_global->DAGChangeType          = DAG_REV; 
					  }                                  
				      nValidMoves++;
#else
//*****/
//**********
#pragma omp critical (ReverseEdges)  //lai, 06-09-2004
					{
                        if (n_totalreverses >= n_reverseslimit) {
                            n_reverseslimit = ((n_totalreverses + 1)&~(REVERSE_EDGES_CHUNK - 1)) + REVERSE_EDGES_CHUNK;  //lai, 06-09-2004
							printf("We update the limit to %d for %d reverses\n",n_reverseslimit,n_totalreverses);
							EDGEREVERSE* newReverseEdges = (EDGEREVERSE *)malloc(n_reverseslimit * sizeof(EDGEREVERSE)); //lai, 06-09-2004
							memcpy(newReverseEdges, ReverseEdges, n_totalreverses * sizeof(EDGEREVERSE)); //lai, 06-09-2004
							delete(ReverseEdges); //lai, 06-09-2004
							ReverseEdges = newReverseEdges; //lai, 06-09-2004
							//printf("promotion completed\n");
						} //lai, 06-09-2004
						ReverseEdges[n_totalreverses].parentNode =  i; //lai, 06-09-2004
                        ReverseEdges[n_totalreverses].childNode  =  j; //lai, 06-09-2004
                        ReverseEdges[n_totalreverses].semi_scoreimprove =  score_improve; //lai, 06-09-2004
						n_totalreverses++; //lai, 06-09-2004
					}
//**********/
#endif
					} //end if newFamily.size() <= m_nMaxFanIn
				} //end if IsValidMove(i, j, DAG_REV)
//#endif
			  } //end if nDirect == 1
		      else if (nDirect == 0)//no edge, try to add it.
              {
				  if(iDAG->IsValidMove(i, j, DAG_ADD, Parents_Buffer_global)) { //lai, 05-17-2004
				    nParrents = iDAG->GetParents(j, newFamily_global); //add an edge
				    newFamily_global[nParrents] = i;
					if((nParrents + 1) <= m_nMaxFanIn) {
					  newFamily_global[nParrents + 1] = j;
					  score_improve = ScoreFamily(nParrents + 2, newFamily_global, indexes_global) - FamilyScore[j]; //lai, 05-17-2004
					  if (score_improve > *max_improve_global) { 
					      *max_improve_global = score_improve;   
					      move_global->originalEdge.startNode =       i; 
					      move_global->originalEdge.endNode   =       j; 
					      move_global->DAGChangeType          = DAG_ADD; 
					  }                                  
				      nValidMoves++;
					} //end if newFamily.size() <= m_nMaxFanIn
                } //end if IsValidMove(i, j, DAG_ADD)
			  } //end if nDirect == 0
#ifdef _OPENMP
			  else { //nDirect == -1, prepare for reverse edge
				if(iDAG->IsValidMove(j, i, DAG_REV, Parents_Buffer_global)) {
				    nParrents = iDAG->GetParents(j, newFamily_global); //add an edge
				    newFamily_global[nParrents] = i;
				    if((nParrents + 1) <= m_nMaxFanIn) {
					  newFamily_global[nParrents + 1] = j;
					  ScoreFamily(nParrents + 2, newFamily_global, indexes_global); //lai, warmup ScoreCache, 05-17-2004
				    } //end if newFamily.size() <= m_nMaxFanIn
				} //end if IsValidMove(i, j, DAG_REV)
			  } //end else nDirect == -1
#endif
			} //end for
		}

#ifdef _OPENMP
//**********
#pragma omp parallel for schedule(static) private(score_improve, score_start, score_end) reduction(+: nValidMoves)
		for(j=0; j<m_nNodes; j++)  {
//#endif
        for( int i = 0; i < n_totalreverses; i ++) {
//#ifdef _OPENMP
			if(j == ReverseEdges[i].parentNode) {
//#endif
                int nParrents = iDAG->GetParents(ReverseEdges[i].parentNode, newFamily_global); //add an edge
				newFamily_global[nParrents] = ReverseEdges[i].childNode;
				newFamily_global[nParrents + 1] = ReverseEdges[i].parentNode;
				score_end = score_improve = ReverseEdges[i].semi_scoreimprove;
				score_start = ScoreFamily(nParrents + 2, newFamily_global, indexes_global) - FamilyScore[ReverseEdges[i].parentNode];
				score_improve += score_start;
				if (score_improve > *max_improve_global) { 
					*max_improve_global = score_improve;   
					*max_start_global   =   score_start;   
					*max_end_global     =     score_end;   
					move_global->originalEdge.startNode =   ReverseEdges[i].parentNode; 
					move_global->originalEdge.endNode   =   ReverseEdges[i].childNode; 
					move_global->DAGChangeType          =	 				DAG_REV; 
				}                                  
				nValidMoves++;
//#ifdef _OPENMP
            }
		}
//#endif
		}
//****************/
/******
#pragma omp parallel for schedule(static) private(score_improve, score_start, score_end, k) reduction(+: nValidMoves)
		for(j=0; j<m_nNodes; j++) {
			for(int i=0; i<m_nNodes; i++)
			{
              int position, nParrents;
		      if(i == j)
			  {
			   continue;
			  }
              int nDirect = iDAG->GetEdgeDirect(i, j);
			  if(nDirect == 1)  // real REV, change, must hit ScoreCache here
			  {
				if(iDAG->IsValidMove(i, j, DAG_REV, Parents_Buffer_global)) {
				   nParrents = iDAG->GetParents(i, newFamily_global); //add an edge
				   newFamily_global[nParrents] = j;
				   if((nParrents + 1) <= m_nMaxFanIn) {
					  newFamily_global[nParrents + 1] = i;
					  score_start = ScoreFamily(nParrents + 2, newFamily_global, indexes_global) - FamilyScore[i]; //lai, 05-11-2004
				      nParrents = iDAG->GetParents(j, newFamily_global);
					  position = nParrents;
					  for( k = 0; k < nParrents; k ++)
						if (newFamily_global[k] == i) {
							position = k;
							break;
						}
					  for(k = position; k < nParrents - 1; k ++)
						newFamily_global[k] = newFamily_global[k + 1];
					  newFamily_global[nParrents - 1] = j;
				      score_end = ScoreFamily(nParrents, newFamily_global, indexes_global) - FamilyScore[j]; //lai, 05-11-2004
					  score_improve = score_start + score_end;
					  if (score_improve > *max_improve_global) { 
					      *max_improve_global = score_improve;   
					      *max_start_global   =   score_start;   
					      *max_end_global     =     score_end;   
					      move_global->originalEdge.startNode =       i; 
					      move_global->originalEdge.endNode   =       j; 
					      move_global->DAGChangeType          = DAG_REV; 
					  }                                  
				      nValidMoves++;
				   } //end if newFamily.size() <= m_nMaxFanIn
				} //end if IsValidMove(i, j, DAG_REV)
			  } //end if nDirect == 1
			} //end for
		}
*********/

//#ifdef 	_OPENMP
		max_improve_global = m_improves_content;
		max_start_global   = m_starts_content;
		max_end_global     = m_ends_content;
		move_global        = moves_content;
        for(k = 1; k < num_threads; k ++) {
			if (m_improves_content[k] > *max_improve_global) {
				max_improve_global = m_improves_content + k;
				max_start_global   = m_starts_content   + k;
				max_end_global     = m_ends_content     + k;
				move_global        = moves_content      + k;
			}
		}
#endif

		if (!step) {
			//VTResume();
                        timer_stop(2);
			printf("end process step 0 at %f\n", (float)(timer_read(2)));
		}

		best_score = *max_improve_global;   //lai, 05-11-2004
		//score_gate = (float)fabs(m_minProgress * init_score);
		score_gate = (float)fabs(1e-5);
		int start = move_global->originalEdge.startNode;
		int end = move_global->originalEdge.endNode;
		//printf("****%4d(%5d %5d)***%8d %5d%4d%12.4f\n", step, position, nValidMoves,
		printf("****%4d(%5d)***%8d %5d%4d%12.4f\n", step, nValidMoves,
			start, end, 
			move_global->DAGChangeType,
			*max_improve_global);  

		if(best_score <= score_gate)
		{
			progress = false;
			break;
		}

		int* ranges = new int[m_nMaxFanIn];  //lai, 07-16-2004
        for(int id=0; id<m_nMaxFanIn; id++) ranges[id] = m_nNodes+1; //lai, 07-16-2004
		
		switch (move_global->DAGChangeType)
		{
		case DAG_DEL :
			if(iDAG->DoMove(start, end, DAG_DEL))
			{ 
				init_score += *max_improve_global;       
				FamilyScore[end] += *max_improve_global; 
				delete m_vNodeScoreCache[end]; //lai, 07-16-2004
				m_vNodeScoreCache[end] = CSparseMatrix<float>::Create(m_nMaxFanIn, ranges, (float)1e38, 0); //lai, 07-16-2004
			}
			break;

		case DAG_ADD :
			if(iDAG->DoMove(start, end, DAG_ADD))
			{
				init_score += *max_improve_global;       
				FamilyScore[end] += *max_improve_global; 
				delete m_vNodeScoreCache[end]; //lai, 07-16-2004
				m_vNodeScoreCache[end] = CSparseMatrix<float>::Create(m_nMaxFanIn, ranges, (float)1e38, 0); //lai, 07-16-2004
			}
			break;

		case DAG_REV :
			if(iDAG->DoMove(start, end, DAG_REV))
			{
				init_score += *max_improve_global;       
				FamilyScore[start] += *max_start_global; 
				FamilyScore[end] += *max_end_global;     
				delete m_vNodeScoreCache[start]; //lai, 07-16-2004
				m_vNodeScoreCache[start] = CSparseMatrix<float>::Create(m_nMaxFanIn, ranges, (float)1e38, 0); //lai, 07-16-2004
				delete m_vNodeScoreCache[end]; //lai, 07-16-2004
				m_vNodeScoreCache[end] = CSparseMatrix<float>::Create(m_nMaxFanIn, ranges, (float)1e38, 0); //lai, 07-16-2004
			}
			break;
		}

        delete []ranges; //lai, 07-16-2004

#ifdef _OPENMP
		delete []moves_content;
		delete []m_starts_content;
		delete []m_ends_content;
		delete []m_improves_content;
#endif

		step++;
	}

	//VTPause();

	delete ReverseEdges; //lai, 06-09-2004
/****
#pragma omp parallel
	{
			delete []newFamily_global;
			delete []indexes_global;
			delete []Parents_Buffer_global;
	}
****/

	*LearnedScore = this->ScoreDAG(iDAG, &FamilyScore);
	//*LearnedDag = iDAG->Clone();
	*LearnedDag = iDAG;
	//delete iDAG;
}

#else

EDGEOP      *move_global;                                            //added by lai, 05-17-2004
float       *max_improve_global, *max_start_global, *max_end_global; //added by lai, 05-17-2004
int         *newFamily_global, *indexes_global, *Parents_Buffer_global;  //added by lai, 05-17-2004
#pragma omp threadprivate(move_global)  //added by lai, 05-17-2004
#pragma omp threadprivate(max_improve_global, max_start_global, max_end_global) //added by lai, 05-17-2004
#pragma omp threadprivate(newFamily_global) //added by lai, 05-17-2004
#pragma omp threadprivate(indexes_global, Parents_Buffer_global) //added by lai, 05-17-2004

#define REVERSE_EDGES_CHUNK  256  //added by lai, 06-09-2004
typedef struct {                  //added by lai, 06-09-2004
  int   parentNode;
  int   childNode;
  float semi_scoreimprove;
} EDGEREVERSE;                    //added by lai, 06-09-2004

typedef struct {                  //added by lai, 06-09-2004
  float    best_score; 
  int  pos_best_score; 
} S_BESTSCORE;					  //added by lai, 06-09-2004

void CMlStaticStructLearnHC::LearnInOneStart(CDAG* InitDag, 
					CDAG** LearnedDag, float* LearnedScore)
{
	int		j, step=0;
	bool	progress = true;
	CDAG*	iDAG = InitDag->Clone();
	//floatVector FamilyScore;
	float* FamilyScore = (float *)malloc(m_nNodes * sizeof(float)); //lai, 06-15-2004
	float init_score = ScoreDAG(iDAG, FamilyScore);	// lai, 06-15-2004
	//float init_score = ScoreDAG(iDAG, &FamilyScore);	// Eric

	int			nPValidMoves,nValidMoves;
	EDGEOPVECTOR vValidMoves;
	float		best_score, score_gate;
	intVector	vAncestor, vDescent;
	intVector	vDiscrete, vContinuous;
	const CNodeType* nt;
	for(j=0; j<m_nNodes; j++)
	{
		nt = m_pGrModel->GetNodeType(j);
		if( nt->IsDiscrete() )
			vDiscrete.push_back(j);
		else
			vContinuous.push_back(j);
	}
	vAncestor.assign(m_vAncestor.begin(), m_vAncestor.end());
	vDescent.assign(m_vDescent.begin(), m_vDescent.end());

#pragma omp parallel
{
	newFamily_global   = new int [m_nMaxFanIn + 1];
	indexes_global     = new int [m_nMaxFanIn];
	Parents_Buffer_global  = new int [m_nMaxFanIn];
}

	int proc_reverseslimit = REVERSE_EDGES_CHUNK;    //lai, 06-09-2004
	int n_procreverses, n_totalreverses;             //lai, 06-09-2004
	n_procreverses  = 0;                             //lai, 06-09-2004
	n_totalreverses = 0;                             //lai, 06-09-2004
    EDGEREVERSE* SplitReverseEdges = (EDGEREVERSE *)malloc(proc_reverseslimit * sizeof(EDGEREVERSE)); //lai, 06-09-2004
	printf("Thread %d compute from node %d to node %d\n", _my_rank, proc_startNode, proc_endNode); //lai, 06-09-2004
	int* nValidReverses   = new int[_num_procs];     //lai, 06-09-2004
	int* nReversePosition = new int[_num_procs];     //lai, 06-09-2004
	int  nTotalReverses;                             //lai, 06-09-2004
	EDGEREVERSE* ValidReverseEdges;                  //lai, 06-09-2004

	//timer_start(3);  //lai, 04-20-2004
    while ( step<m_maxIterIPF && progress )
	//while ( step<m_nSteps && progress )
	{
		nPValidMoves = 0;  //lai, 06-15-2004
		n_procreverses  = 0;                             //lai, 06-09-2004
		float score_improve, score_start, score_end;  //lai, 05-11-2004
		int k;

#ifndef _OPENMP
		EDGEOP  move_content;
		move_content.originalEdge.startNode =       0;
        move_content.originalEdge.endNode   =       0;
		move_content.DAGChangeType          = DAG_ADD;
		float m_start_content, m_end_content;         //lai, 05-11-2004
		float m_improve_content = 0.0f;               //lai, 05-11-2004
		move_global = &move_content;
        max_improve_global = &m_improve_content;
		max_start_global   = &m_start_content;
		max_end_global     = &m_end_content;
#else
		int num_threads = omp_get_max_threads();
		EDGEOP *moves_content = new EDGEOP[num_threads];
		float *m_starts_content = new float[num_threads];
		float *m_ends_content = new float[num_threads];
		float *m_improves_content = new float[num_threads];
        for(k = 0; k < num_threads; k ++) {
			moves_content[k].originalEdge.startNode =       0;
			moves_content[k].originalEdge.endNode   =       0;
			moves_content[k].DAGChangeType          = DAG_ADD;
			m_improves_content[k] = 0.0f;
		}
#pragma omp parallel
		{
			move_global = moves_content + omp_get_thread_num();
			max_improve_global = m_improves_content + omp_get_thread_num();
			max_start_global   = m_starts_content + omp_get_thread_num();
			max_end_global     = m_ends_content + omp_get_thread_num();
		}
#endif

		//printf("\nbegin process moves at %f\n", (float)(clock()-start_GetMoves)/CLOCKS_PER_SEC);

#pragma omp parallel for schedule(static) private(score_improve, k) reduction(+: nPValidMoves)
		for(j=proc_startNode; j<proc_endNode; j++) {  //lai, 06-09-2004
		//for(j=0; j<m_nNodes; j++) {
		    for(int i=0; i<m_nNodes; i++)
			{
			  //newFamily->clear();     //lai, 05-17-2004
              int position, nParrents;
		      if(i == j)
			  {
			   continue;
			  }
              int nDirect = iDAG->GetEdgeDirect(i, j);
			  if(nDirect == 1)  // try remove, change
			  {
			    //if((!bCheck || CheckValid(*pvAncesstorVector, *pvDescendantsVector, eOP)) && (!bCheck2 || CheckValid2(*pvNotParents, *pvNotChild, eOP)))
				nParrents = iDAG->GetParents(j, newFamily_global); //lai, 05-17-2004
                position = nParrents;
                for( k = 0; k < nParrents; k ++)
					if (newFamily_global[k] == i) {
						position = k;
						break;
					}
				for(k = position; k < nParrents - 1; k ++)
					newFamily_global[k] = newFamily_global[k + 1];
				newFamily_global[nParrents - 1] = j;
				score_improve = ScoreFamily(nParrents, newFamily_global, indexes_global) - FamilyScore[j]; //lai, 05-17-2004
				if (score_improve > *max_improve_global) { 
					*max_improve_global = score_improve;   
					move_global->originalEdge.startNode =       i; 
					move_global->originalEdge.endNode   =       j; 
					move_global->DAGChangeType          = DAG_DEL; 
				}                                  
				nPValidMoves++;

				if(iDAG->IsValidMove(i, j, DAG_REV, Parents_Buffer_global)) { //lai, 05-17-2004
				    nParrents = iDAG->GetParents(i, newFamily_global); //add an edge
					newFamily_global[nParrents] = j;
					if((nParrents + 1) <= m_nMaxFanIn) { //lai, 05-17-2004
#pragma omp critical (SplitReverseEdges)  //lai, 06-09-2004
					{
                        if (n_procreverses >= proc_reverseslimit) {
                            proc_reverseslimit = ((n_procreverses + 1)&~(REVERSE_EDGES_CHUNK - 1)) + REVERSE_EDGES_CHUNK;  //lai, 06-09-2004
							printf("We update the limit to %d for %d reverses\n",proc_reverseslimit,n_procreverses);
							EDGEREVERSE* newSplitReverseEdges = (EDGEREVERSE *)malloc(proc_reverseslimit * sizeof(EDGEREVERSE)); //lai, 06-09-2004
							memcpy(newSplitReverseEdges, SplitReverseEdges, n_procreverses * sizeof(EDGEREVERSE)); //lai, 06-09-2004
							delete(SplitReverseEdges); //lai, 06-09-2004
							SplitReverseEdges = newSplitReverseEdges; //lai, 06-09-2004
							//printf("promotion completed\n");
						} //lai, 06-09-2004
						SplitReverseEdges[n_procreverses].parentNode =  i; //lai, 06-09-2004
                        SplitReverseEdges[n_procreverses].childNode  =  j; //lai, 06-09-2004
                        SplitReverseEdges[n_procreverses].semi_scoreimprove =  score_improve; //lai, 06-09-2004
						n_procreverses++; //lai, 06-09-2004
					}
					} //end if newFamily.size() <= m_nMaxFanIn
				} //end if IsValidMove(i, j, DAG_REV)
			  } //end if nDirect == 1
		      else if (nDirect == 0)//no edge, try to add it.
              {
				  if(iDAG->IsValidMove(i, j, DAG_ADD, Parents_Buffer_global)) { //lai, 05-17-2004
				    nParrents = iDAG->GetParents(j, newFamily_global); //add an edge
				    newFamily_global[nParrents] = i;
					if((nParrents + 1) <= m_nMaxFanIn) {
					  newFamily_global[nParrents + 1] = j;
					  score_improve = ScoreFamily(nParrents + 2, newFamily_global, indexes_global) - FamilyScore[j]; //lai, 05-17-2004
					  if (score_improve > *max_improve_global) { 
					      *max_improve_global = score_improve;   
					      move_global->originalEdge.startNode =       i; 
					      move_global->originalEdge.endNode   =       j; 
					      move_global->DAGChangeType          = DAG_ADD; 
					  }                                  
				      nPValidMoves++;
					} //end if newFamily.size() <= m_nMaxFanIn
                } //end if IsValidMove(i, j, DAG_ADD)
			  } //end if nDirect == 0
			  else { //nDirect == -1, prepare for reverse edge
				if(iDAG->IsValidMove(j, i, DAG_REV, Parents_Buffer_global)) {
				    nParrents = iDAG->GetParents(j, newFamily_global); //add an edge
				    newFamily_global[nParrents] = i;
				    if((nParrents + 1) <= m_nMaxFanIn) {
					  newFamily_global[nParrents + 1] = j;
					  ScoreFamily(nParrents + 2, newFamily_global, indexes_global); //lai, warmup ScoreCache, 05-17-2004
				    } //end if newFamily.size() <= m_nMaxFanIn
				} //end if IsValidMove(i, j, DAG_REV)
			  } //end else nDirect == -1
			} //end for
		}

        //printf("This node create %d reverseEdges and %d add/remove edges to handle\n",n_procreverses, nPValidMoves);  //lai, 06-15-2004
		nValidReverses[_my_rank] = n_procreverses;
		MPI_Status status;
        for(j=0; j<_num_procs; j++)  MPI_CALL(MPI_Bcast(&(nValidReverses[j]), 1, MPI_INT, j, MPI_COMM_WORLD));
		nReversePosition[0] = 0;
		for(j=1; j<_num_procs; j++)  nReversePosition[j] = nValidReverses[j - 1] + nReversePosition[j - 1];
		nTotalReverses = nReversePosition[_num_procs - 1] + nValidReverses[_num_procs - 1];
        //printf("Totally we have %d reverseEdges to handle\n",nTotalReverses);   //lai, 06-15-2004
		ValidReverseEdges = (EDGEREVERSE *)malloc(nTotalReverses * sizeof(EDGEREVERSE)); //lai, 06-09-2004
		memcpy(&(ValidReverseEdges[nReversePosition[_my_rank]]), SplitReverseEdges, n_procreverses * sizeof(EDGEREVERSE)); //lai, 06-10-2004
        for(j=0; j<_num_procs; j++)  
			MPI_CALL(MPI_Bcast(&(ValidReverseEdges[nReversePosition[j]]), nValidReverses[j] * sizeof(EDGEREVERSE), MPI_CHAR, j, MPI_COMM_WORLD));
        
        for( j = 0; j < nTotalReverses; j ++) {
          if ((ValidReverseEdges[j].parentNode >= proc_startNode) && (ValidReverseEdges[j].parentNode < proc_endNode)) {
                int nParrents = iDAG->GetParents(ValidReverseEdges[j].parentNode, newFamily_global); //add an edge
				newFamily_global[nParrents] = ValidReverseEdges[j].childNode;
				newFamily_global[nParrents + 1] = ValidReverseEdges[j].parentNode;
				score_end = score_improve = ValidReverseEdges[j].semi_scoreimprove;
				score_start = ScoreFamily(nParrents + 2, newFamily_global, indexes_global) - FamilyScore[ValidReverseEdges[j].parentNode];
				score_improve += score_start;
				if (score_improve > *max_improve_global) { 
					*max_improve_global = score_improve;   
					*max_start_global   =   score_start;   
					*max_end_global     =     score_end;   
					move_global->originalEdge.startNode =   ValidReverseEdges[j].parentNode; 
					move_global->originalEdge.endNode   =    ValidReverseEdges[j].childNode; 
					move_global->DAGChangeType          =							DAG_REV; 
				}                                  
				nPValidMoves++;
		  }
		}

#ifdef _OPENMP
		max_improve_global = m_improves_content;
		max_start_global   = m_starts_content;
		max_end_global     = m_ends_content;
		move_global        = moves_content;
        for(k = 1; k < num_threads; k ++) {
			if (m_improves_content[k] > *max_improve_global) {
				max_improve_global = m_improves_content + k;
				max_start_global   = m_starts_content   + k;
				max_end_global     = m_ends_content     + k;
				move_global        = moves_content      + k;
			}
		}
#endif

		//printf("end process moves at %f\n", (float)(clock()-start_GetMoves)/CLOCKS_PER_SEC);

		//printf("****%4d(%5d)***%8d %5d%4d%12.4f\n", step, nPValidMoves,
		//	move_global->originalEdge.startNode, move_global->originalEdge.endNode, 
		//	move_global->DAGChangeType, *max_improve_global);  
        S_BESTSCORE choose_best, choose_best_result;
		choose_best.best_score     = *max_improve_global;
        choose_best.pos_best_score =   _my_rank;
		MPI_CALL(MPI_Allreduce( &choose_best, &choose_best_result, 1, MPI_FLOAT_INT, MPI_MAXLOC, MPI_COMM_WORLD ));
        best_score = *max_improve_global = choose_best_result.best_score;
		MPI_CALL(MPI_Bcast(move_global, sizeof(EDGEOP), MPI_CHAR, choose_best_result.pos_best_score, MPI_COMM_WORLD));
		MPI_CALL(MPI_Allreduce( &nPValidMoves, &nValidMoves, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD ));
        if(move_global->DAGChangeType == DAG_REV) {
			MPI_CALL(MPI_Bcast(max_start_global, 1, MPI_FLOAT, choose_best_result.pos_best_score, MPI_COMM_WORLD));
			MPI_CALL(MPI_Bcast(max_end_global, 1, MPI_FLOAT, choose_best_result.pos_best_score, MPI_COMM_WORLD));
		}
		score_gate = (float)fabs(1e-5);
		int start = move_global->originalEdge.startNode;
		int end = move_global->originalEdge.endNode;
		//printf("****%4d(%5d %5d)***%8d %5d%4d%12.4f\n", step, position, nValidMoves,
		printf("****%4d(%5d)***%8d %5d%4d%12.4f\n", step, nValidMoves,
			start, end, 
			move_global->DAGChangeType,
			*max_improve_global);  

		if(best_score <= score_gate)
		{
			progress = false;
			break;
		}

		switch (move_global->DAGChangeType)
		{
		case DAG_DEL :
			if(iDAG->DoMove(start, end, DAG_DEL))
			{ 
				init_score += *max_improve_global;       
				FamilyScore[end] += *max_improve_global; 
			}
			break;

		case DAG_ADD :
			if(iDAG->DoMove(start, end, DAG_ADD))
			{
				init_score += *max_improve_global;       
				FamilyScore[end] += *max_improve_global; 
			}
			break;

		case DAG_REV :
			if(iDAG->DoMove(start, end, DAG_REV))
			{
				init_score += *max_improve_global;       
				FamilyScore[start] += *max_start_global; 
				FamilyScore[end] += *max_end_global;     
			}
			break;
		}

		delete ValidReverseEdges;  //added by lai, 06-09-2004

#ifdef _OPENMP
		delete []moves_content;
		delete []m_starts_content;
		delete []m_ends_content;
		delete []m_improves_content;
#endif

		step++;
	}

    delete SplitReverseEdges; //lai, 06-09-2004
	delete[] nValidReverses;     //lai, 06-09-2004
	delete[] nReversePosition;     //lai, 06-09-2004

/****
#pragma omp parallel
	{
			delete []newFamily_global;
			delete []indexes_global;
			delete []Parents_Buffer_global;
	}
****/

	//*LearnedScore = this->ScoreDAG(iDAG, &FamilyScore);
	//*LearnedScore = this->ScoreDAG(iDAG, FamilyScore);
	*LearnedScore = init_score;  //added by lai, 06-15-2004
	delete(FamilyScore);         //added by lai, 06-15-2004
	//*LearnedDag = iDAG->Clone();
	*LearnedDag = iDAG;
	//delete iDAG;
}

#endif

