/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                INTEL CORPORATION PROPRIETARY INFORMATION                //
//   This software is supplied under the terms of a license agreement or   //
//  nondisclosure agreement with Intel Corporation and may not be copied   //
//   or disclosed except in accordance with the terms of that agreement.   //
//       Copyright (c) 2003 Intel Corporation. All Rights Reserved.        //
//                                                                         //
//  File:      pnlMlStaticStructLearnHC.cpp                                //
//                                                                         //
//  Purpose:   Implementation of the CMlStaticStructLearnHC class          //
//                                                                         //
//  Author(s):                                                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////
#include "pnlConfig.hpp"
#include "pnlMlStaticStructLearnHC.hpp"

#include "mpiWrap.h" 

PNL_USING

#define _OPT

CMlStaticStructLearnHC::CMlStaticStructLearnHC( CStaticGraphicalModel* pGrModel, 
						  ELearningTypes LearnType, 
						  EOptimizeTypes AlgorithmType, 
						  EScoreFunTypes ScoreType, 
						  int nMaxFanIn,
						  intVector& vAncestor,
						  intVector& vDescent,
						  int nRestarts ):
							CMlStaticStructLearn( pGrModel, LearnType, 
								                   AlgorithmType,ScoreType,
												   nMaxFanIn,vAncestor,vDescent )
{
	m_nRestarts = nRestarts;
	m_minProgress = (float)1e-4;
	m_nSteps = 100;
}

CMlStaticStructLearnHC* CMlStaticStructLearnHC::Create( CStaticGraphicalModel* pGrModel, 
						  ELearningTypes LearnType, 
						  EOptimizeTypes AlgorithmType, 
						  EScoreFunTypes ScoreType, 
						  int nMaxFanIn,
						  intVector& vAncestor,
						  intVector& vDescent,
						  int nRestarts )
{
	CMlStaticStructLearnHC* ret = new CMlStaticStructLearnHC( pGrModel, LearnType, 
								                   AlgorithmType,ScoreType,
												   nMaxFanIn, vAncestor,
												   vDescent, nRestarts );
	return ret;
}

void CMlStaticStructLearnHC::SetMinProgress(float minProgress)
{
	m_minProgress = minProgress;
}

void CMlStaticStructLearnHC::SetMaxSteps(int steps)
{
	m_nSteps = steps;
}

void CMlStaticStructLearnHC::Learn()
{
	if(m_Vector_pEvidences.size() == 0)
		PNL_THROW(CInconsistentState, "should set the data first");
	
	CGraph* iGraph = m_pGrModel->GetGraph();
	CDAG* iDAG = CDAG::Create(*iGraph);
	CDAG* pDAG;
	CDAG* pBestDAG = NULL;
	float BestScore = (float)-1e37;
	int irestarts = m_nRestarts;
	int i, istart;
	float score;

	for(istart=0; istart<irestarts; istart++)
	{
		if(istart>0)
		{
			delete iDAG;
			intVector	vDiscrete, vContinuous;
			const CNodeType* nt;
			for(i=0; i<m_nNodes; i++)
			{
				nt = m_pGrModel->GetNodeType(i);
				if( nt->IsDiscrete() )
					vDiscrete.push_back(i);
				else
					vContinuous.push_back(i);
			}
			iDAG = CDAG::RandomCreateADAG(m_nNodes, m_nMaxFanIn, vDiscrete, vContinuous);
		}
		LearnInOneStart(iDAG, &pDAG, &score);
		if(score > BestScore)
		{
			delete pBestDAG;
			pBestDAG = pDAG->Clone();
			BestScore = score;
		}
		delete pDAG;
	}
	delete iDAG;
	m_pResultDAG = pBestDAG->Clone();
	m_critValue.push_back(BestScore);
	delete pBestDAG;
}

#ifndef _MPI

void CMlStaticStructLearnHC::LearnInOneStart(CDAG* InitDag, 
					CDAG** LearnedDag, float* LearnedScore)
{
	int		i, j, step=0;
	bool	progress = true;
	CDAG*	iDAG = InitDag->Clone();
	floatVector FamilyScore;
	float init_score = ScoreDAG(iDAG, &FamilyScore);	// Eric

	int			nValidMoves;
	EDGEOPVECTOR vValidMoves;
	EDGEOP      move;
	float		best_score, score_gate;
	intVector   newFamily;
	intVector	vAncestor, vDescent;
	intVector	vDiscrete, vContinuous;
	int			start, end, position;
	const CNodeType* nt;
	for(i=0; i<m_nNodes; i++)
	{
		nt = m_pGrModel->GetNodeType(i);
		if( nt->IsDiscrete() )
			vDiscrete.push_back(i);
		else
			vContinuous.push_back(i);
	}
	vAncestor.assign(m_vAncestor.begin(), m_vAncestor.end());
	vDescent.assign(m_vDescent.begin(), m_vDescent.end());
	
	while ( step<m_nSteps && progress )
	{
		iDAG->GetAllValidMove(&vValidMoves,&vDiscrete, &vContinuous, &vDescent, &vAncestor);
		nValidMoves = vValidMoves.size();
		floatVector scores_improve(nValidMoves);
		floatVector score_start(nValidMoves);
		floatVector score_end(nValidMoves);
		for(i=0; i<nValidMoves; i++) 
		{
			scores_improve[i] = 0.0f;
			newFamily.clear();
			move = vValidMoves[i];
			switch (move.DAGChangeType)
			{
			case DAG_DEL : 
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(end, &newFamily);
				newFamily.push_back(end);
				position = std::find(newFamily.begin(), newFamily.end(), start)
					       - newFamily.begin();
				newFamily.erase(newFamily.begin()+position);

				scores_improve[i] = ScoreFamily(newFamily) - FamilyScore[end];
				break;

			case DAG_ADD :
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(end, &newFamily);
#ifdef _OPT  //added by lai, 03-30-2004
				newFamily.push_back(start);
#else
				position = newFamily.size();
				for(j=0; j<newFamily.size(); j++)
				{
					if(start<newFamily[j])
					{
						position = j;
						break;
					}
				}
				if(position == newFamily.size())
					newFamily.push_back(start);
				else
					newFamily.insert(newFamily.begin()+position, start);
#endif
				newFamily.push_back(end);
				if(newFamily.size() > (m_nMaxFanIn+1))
					break;
				scores_improve[i] = ScoreFamily(newFamily) - FamilyScore[end];
				break;

			case DAG_REV :
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(start, &newFamily); //add an edge
#ifdef _OPT  //added by lai, 03-30-2004
				newFamily.push_back(end);
#else
				position = newFamily.size();
				for(j=0; j<newFamily.size(); j++)
				{
					if(end<newFamily[j])
					{
						position = j;
						break;
					}
				}
				if(position == newFamily.size())
					newFamily.push_back(end);
				else
					newFamily.insert(newFamily.begin()+position, end);		
#endif
				newFamily.push_back(start);
				if(newFamily.size() > (m_nMaxFanIn+1))
					break;
				scores_improve[i] = ScoreFamily(newFamily) - FamilyScore[start];
				score_start[i] = scores_improve[i];

				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(end, &newFamily);
				newFamily.push_back(end);
				position = std::find(newFamily.begin(), newFamily.end(), start)
					       - newFamily.begin();
				newFamily.erase(newFamily.begin()+position);

				score_end[i] = ScoreFamily(newFamily) - FamilyScore[end];
                scores_improve[i] += score_end[i];
				break;
			}
		}

		position = std::max_element(scores_improve.begin(), scores_improve.begin()+nValidMoves) - scores_improve.begin();
		move = vValidMoves[position];
		start = move.originalEdge.startNode;
		end = move.originalEdge.endNode;
		printf("\n****%4d(%4d %4d)***%8d%4d%4d%12.4f", step, position, nValidMoves,
			start, end, 
			vValidMoves[position].DAGChangeType, 
			scores_improve[position]); 

		best_score = scores_improve[position];
		score_gate = (float)fabs(m_minProgress * init_score);
		if(best_score <= score_gate)
		{
			progress = false;
			break;
		}

		switch (move.DAGChangeType)
		{
		case DAG_DEL :
			if(iDAG->DoMove(start, end, DAG_DEL))
			{ 
				init_score += scores_improve[position];
				FamilyScore[end] += scores_improve[position];
			}
			break;

		case DAG_ADD :
			if(iDAG->DoMove(start, end, DAG_ADD))
			{
				init_score += scores_improve[position];
				FamilyScore[end] += scores_improve[position];
			}
			break;

		case DAG_REV :
			if(iDAG->DoMove(start, end, DAG_REV))
			{
				init_score += scores_improve[position];
				FamilyScore[start] += score_start[position];
				FamilyScore[end] += score_end[position];
			}
			break;
		}

		step++;
	}

	*LearnedScore = this->ScoreDAG(iDAG, &FamilyScore);
	*LearnedDag = iDAG->Clone();
	delete iDAG;
}

#else


#define  PRE_ALLOC_CHUCK 4096

float *p_Shared_MirrorBuf, *p_Shared_Buf;
int   *nSharedBufLen;
int   nTotalBufLen;
int   nInteration;
int   bParallel;

int ScoreCacheExisted( intVector vFamily, int nMaxFamily, CSparseMatrix<float> *pFamilyScoreCache, int *indexes)
{
  int i;
  int nFamily = vFamily.size();
  
  for( i = 0; i < nMaxFamily; i++ )
    indexes[i] = 0;

  for( i = 0; i < nFamily; i++ )
    indexes[i] = vFamily[i] + 1;
 
  return pFamilyScoreCache->IsExistingElement(indexes);
}

void CMlStaticStructLearnHC::LearnInOneStart(CDAG* InitDag, 
					CDAG** LearnedDag, float* LearnedScore)
{
	int	i, j, step=0;
	bool	progress = true;
		
	// Init memory for exchange buffer
	p_Shared_MirrorBuf = new float [m_nNodes * PRE_ALLOC_CHUCK];
	p_Shared_Buf = new float [m_nNodes * PRE_ALLOC_CHUCK];
	nSharedBufLen = new int [m_nNodes*m_nNodes];
	nTotalBufLen = 0;
	nInteration = 0;
	bParallel = 1;

	CDAG*	iDAG = InitDag->Clone();
	floatVector FamilyScore;
	float init_score; 
	int			nValidMoves;
	EDGEOPVECTOR vValidMoves;
	EDGEOP      move;
	float		best_score, score_gate;
	intVector   newFamily;
	intVector	vAncestor, vDescent;
	intVector	vDiscrete, vContinuous;
	int			start, end, position;
	const CNodeType* nt;
	for(i=0; i<m_nNodes; i++)
	{
		nt = m_pGrModel->GetNodeType(i);
		if( nt->IsDiscrete() )
			vDiscrete.push_back(i);
		else
			vContinuous.push_back(i);
	}
	vAncestor.assign(m_vAncestor.begin(), m_vAncestor.end());
	vDescent.assign(m_vDescent.begin(), m_vDescent.end());

	while ( step<m_nSteps && progress )
	{
		iDAG->GetAllValidMove(&vValidMoves,&vDiscrete, &vContinuous, &vDescent, &vAncestor);
		nValidMoves = vValidMoves.size();
		floatVector scores_improve(nValidMoves+m_nNodes);
		floatVector score_start(nValidMoves+m_nNodes);
		floatVector score_end(nValidMoves+m_nNodes);

		int nNonExistScore = 0, nMaxFamily = m_nMaxFanIn + 1;
		int *pIndexChain = new int [(nValidMoves+m_nNodes)*nMaxFamily];
		int *pIndexNum = new int [(nValidMoves+m_nNodes) ];
		int *pIndexes = new int [nMaxFamily];
		
		if (step == 0)
		{
			for(int i=0; i<m_nNodes; i++)
			{
				newFamily.clear();
				iDAG->GetParents(i, &newFamily);
				newFamily.push_back(i);
				
				if ( !ScoreCacheExisted(newFamily, nMaxFamily, m_pFamilyScoreCache, pIndexes))
				{
					for (j = 0; j < nMaxFamily; j++)
						pIndexChain[nNonExistScore*nMaxFamily+j] = pIndexes[j];
					pIndexNum[nNonExistScore] = newFamily.size();
					nNonExistScore++;
				}
			}
		}

		for(i=0; i<nValidMoves; i++) 
		{
			scores_improve[i] = 0.0f;
			newFamily.clear();
			move = vValidMoves[i];
			switch (move.DAGChangeType)
			{
			case DAG_DEL : 
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(end, &newFamily);
				newFamily.push_back(end);
				position = std::find(newFamily.begin(), newFamily.end(), start)
					- newFamily.begin();
				newFamily.erase(newFamily.begin()+position);
				
				if ( !ScoreCacheExisted(newFamily, nMaxFamily, m_pFamilyScoreCache, pIndexes))
				{
					for (j = 0; j < nMaxFamily; j++)
						pIndexChain[nNonExistScore*nMaxFamily+j] = pIndexes[j];
					pIndexNum[nNonExistScore] = newFamily.size();
					nNonExistScore++;
				}
				break;
				
			case DAG_ADD :
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(end, &newFamily);
#ifdef _OPT   //added by lai, 03-30-2004
				newFamily.push_back(start);
#else
				position = newFamily.size();
				for(j=0; j<newFamily.size(); j++)
				{
					if(start<newFamily[j])
					{
						position = j;
						break;
					}
				}
				if(position == newFamily.size())
					newFamily.push_back(start);
				else
					newFamily.insert(newFamily.begin()+position, start);		
#endif
				newFamily.push_back(end);
				if(newFamily.size() > (m_nMaxFanIn+1))
					break;
				
				if ( !ScoreCacheExisted(newFamily, nMaxFamily, m_pFamilyScoreCache, pIndexes))
				{
					for (j = 0; j < nMaxFamily; j++)
						pIndexChain[nNonExistScore*nMaxFamily+j] = pIndexes[j];
					pIndexNum[nNonExistScore] = newFamily.size();
					nNonExistScore++;
				}
				
				break;
				
			case DAG_REV :
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(start, &newFamily); //add an edge
#ifdef _OPT   //added by lai, 03-30-2004
				newFamily.push_back(end);
#else
				position = newFamily.size();
				for(j=0; j<newFamily.size(); j++)
				{
					if(end<newFamily[j])
					{
						position = j;
						break;
					}
				}
				if(position == newFamily.size())
					newFamily.push_back(end);
				else
					newFamily.insert(newFamily.begin()+position, end);		
#endif	
				newFamily.push_back(start);
				if(newFamily.size() > (m_nMaxFanIn+1))
					break;
				
				if ( !ScoreCacheExisted(newFamily, nMaxFamily, m_pFamilyScoreCache, pIndexes))
				{
					for (j = 0; j < nMaxFamily; j++)
						pIndexChain[nNonExistScore*nMaxFamily+j] = pIndexes[j];
					pIndexNum[nNonExistScore] = newFamily.size();
					nNonExistScore++;
				}
				break;
			}
		}

		// Reinit
		bParallel = 1;
		nTotalBufLen = 0;
		nInteration = 0;

		// Data domain decomposition
		for (i = 0; i < nNonExistScore; i++)
		{
			newFamily.clear();
			for (j = 0; j < pIndexNum[i]; j++ )
				newFamily.push_back(pIndexChain[i*nMaxFamily+j]-1);
			
			ComputeFamilyScore(newFamily);
		}

		// Exchange middle layer data
		MPI_CALL( MPI_Allreduce(p_Shared_MirrorBuf, p_Shared_Buf, nTotalBufLen, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD) );
		
		bParallel = 0;
		nTotalBufLen = 0;
		nInteration = 0;
	
		for (i = 0; i < nNonExistScore; i++)
		{
			newFamily.clear();
			for (j = 0; j < pIndexNum[i]; j++ )
				newFamily.push_back(pIndexChain[i*nMaxFamily+j]-1);
			float score = ComputeFamilyScore(newFamily);
			memcpy(pIndexes, pIndexChain+i*nMaxFamily, sizeof(int)*nMaxFamily);
		
			m_pFamilyScoreCache->SetElementByIndexes(score, pIndexes);
			nTotalBufLen += nSharedBufLen[nInteration];
			nInteration ++;
		}

		delete []pIndexes;
		delete []pIndexChain;
		delete []pIndexNum;
		
		// First step DAG score computation
		if (step == 0)
		{
			init_score = ScoreDAG(iDAG, &FamilyScore);
		}

		for(i=0; i<nValidMoves; i++) 
		{
			scores_improve[i] = 0.0f;
			newFamily.clear();
			move = vValidMoves[i];
			switch (move.DAGChangeType)
			{
			case DAG_DEL : 
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(end, &newFamily);
				newFamily.push_back(end);
				position = std::find(newFamily.begin(), newFamily.end(), start)
					- newFamily.begin();
				newFamily.erase(newFamily.begin()+position);
				
				scores_improve[i] = ScoreFamily(newFamily) - FamilyScore[end];
				break;
				
			case DAG_ADD :
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(end, &newFamily);
#ifdef _OPT   //added by lai, 03-30-2004
				newFamily.push_back(start);
#else
				position = newFamily.size();
				for(j=0; j<newFamily.size(); j++)
				{
					if(start<newFamily[j])
					{
						position = j;
						break;
					}
				}
				if(position == newFamily.size())
					newFamily.push_back(start);
				else
					newFamily.insert(newFamily.begin()+position, start);		
#endif
		
				newFamily.push_back(end);
				if(newFamily.size() > (m_nMaxFanIn+1))
					break;
				
				scores_improve[i] = ScoreFamily(newFamily) - FamilyScore[end];
				break;
				
			case DAG_REV :
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(start, &newFamily); //add an edge
#ifdef _OPT   //added by lai, 03-30-2004
				newFamily.push_back(end);
#else
				position = newFamily.size();
				for(j=0; j<newFamily.size(); j++)
				{
					if(end<newFamily[j])
					{
						position = j;
						break;
					}
				}
				if(position == newFamily.size())
					newFamily.push_back(end);
				else
					newFamily.insert(newFamily.begin()+position, end);		
#endif	
	
				newFamily.push_back(start);
				if(newFamily.size() > (m_nMaxFanIn+1))
					break;
				
				scores_improve[i] = ScoreFamily(newFamily) - FamilyScore[start];
				
				score_start[i] = scores_improve[i];
				
				start = move.originalEdge.startNode;
				end = move.originalEdge.endNode;
				iDAG->GetParents(end, &newFamily);
				newFamily.push_back(end);
				position = std::find(newFamily.begin(), newFamily.end(), start)
					- newFamily.begin();
				newFamily.erase(newFamily.begin()+position);
				score_end[i] = ScoreFamily(newFamily) - FamilyScore[end];
				scores_improve[i] += score_end[i];
				break;
			}
		}
		
		position = std::max_element(scores_improve.begin(), scores_improve.begin()+nValidMoves) - scores_improve.begin();
		best_score = scores_improve[position];
		score_gate = (float)fabs(m_minProgress * init_score);

		if(best_score <= score_gate)
		{
			progress = false;
			break;
		}

		move = vValidMoves[position];
		start = move.originalEdge.startNode;
		end = move.originalEdge.endNode;

		// Verifications
		printf("\n****%4d(%4d)***%8d%4d%4d%12.4f", step, nNonExistScore,
			start, end, 
			vValidMoves[position].DAGChangeType, 
			scores_improve[position]);

		switch (move.DAGChangeType)
		{
		case DAG_DEL :
			if(iDAG->DoMove(start, end, DAG_DEL))
			{ 
				init_score += scores_improve[position];
				FamilyScore[end] += scores_improve[position];
			}
			break;

		case DAG_ADD :
			if(iDAG->DoMove(start, end, DAG_ADD))
			{
				init_score += scores_improve[position];
				FamilyScore[end] += scores_improve[position];
			}
			break;

		case DAG_REV :
			if(iDAG->DoMove(start, end, DAG_REV))
			{
				init_score += scores_improve[position];
				FamilyScore[start] += score_start[position];
				FamilyScore[end] += score_end[position];
			}
			break;
		}
		step++;	
	}

	// Free the exchange buffer
	delete []p_Shared_MirrorBuf;
	delete []p_Shared_Buf;
	delete []nSharedBufLen;

	*LearnedScore = this->ScoreDAG(iDAG, &FamilyScore);
	*LearnedDag = iDAG->Clone();

	delete iDAG;
}
#endif




